#include <upc_relaxed.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <string.h>

#define ITER 100
int CACHE_LINE;
int INTERVAL;
int N;
int sync_type;


shared[96] int *global_arr_ptr, *start;
int *private_arr;
upc_lock_t     *lock;


/* 
   This benchmark is to be run with 4 threads
   All the threads performs updates on shared array located on Thread 0
   The update access straddle cache lines. ie each update access two adjacent cache lines 
   Based on the order in which update takes place, there could be the following scenarios
   (i) all the updates happen and are visible in a certain order
   (ii) due to the order of updates, some updates are not visible
   (iii) Some updates are partially reflected -- (WRONG UPDATES)
*/



int bulk_access(int sync_type)
{
	int i,j;
        int num, offset;
	num = 1.5*CACHE_LINE;
        offset = MYTHREAD * ((1.5 * CACHE_LINE) + INTERVAL); //* sizeof(int);

        if (sync_type == 0){ /* use lock/unlock sync */
		for(j=0;j<ITER;j++)
		{	
			upc_lock(lock);
			upc_memget(private_arr,  (global_arr_ptr + offset), num*sizeof(int));
			for (i = 0; i < num; i++) {
				private_arr[i] = private_arr[i] + 1;
			}	
			upc_memput((global_arr_ptr + offset), private_arr, num*sizeof(int));
			upc_unlock(lock);
        	}
        }

        else if (sync_type == 1){ /* no locks */
                for(j=0;j<ITER;j++)
                { 
                        upc_memget(&private_arr[0], global_arr_ptr + offset, num*sizeof(int));
                        for (i = 0; i < num; i++) {
                                private_arr[i] = private_arr[i] + 1;
                        }
                        upc_memput(global_arr_ptr + offset, &private_arr[0], num*sizeof(int));
                }
        }
        return NULL;
}

/* test the different ways of performing updates */
void run_update_test()
{
	bulk_access(sync_type);
        //direct_access(sync_type);

}

/* test the correctness of the updates */
int verify_updates()
{
	int i;
        int nerr=0;
    	for(i=0;i<N;i++)
	{ 
		if (i < 1.5*CACHE_LINE){
			if (global_arr_ptr[i] != ITER) nerr++;}
		else if (i >= 1.5*CACHE_LINE && i < (1.5*CACHE_LINE + INTERVAL)){
			if (global_arr_ptr[i] != 0) nerr++;}
		else if (i >= 1.5*CACHE_LINE+INTERVAL && i < 3*CACHE_LINE+INTERVAL){
			if (global_arr_ptr[i] != 10000+ITER) nerr++;}
		else if (i >= 3*CACHE_LINE+INTERVAL && i < 3*CACHE_LINE+2*INTERVAL){
			if (global_arr_ptr[i] != 10000) nerr++;}
		else if (i >= 3*CACHE_LINE+2*INTERVAL && i < 4.5*CACHE_LINE + 2*INTERVAL){
			if (global_arr_ptr[i] != 20000+ITER) nerr++;}
		else if (i >= 4.5*CACHE_LINE + 2*INTERVAL && i < 4.5*CACHE_LINE + 3*INTERVAL){
			if (global_arr_ptr[i] != 20000) nerr++;}
		else if (i >= 4.5*CACHE_LINE+3*INTERVAL && i < 6*CACHE_LINE + 3*INTERVAL){
			if (global_arr_ptr[i] != 30000+ITER) nerr++;}
		else if (i >= 6*CACHE_LINE + 3*INTERVAL && i < N){
			if (global_arr_ptr[i] != 30000) nerr++;}

	}

	/* output the num of errors */
	if(nerr == 0)
		printf("NO ERRORS\n");
	else
		printf("RESULTED IN %d ERRORS\n",nerr);

       /* for (i=0;i<N;i++)
        printf("%d\t-", global_arr_ptr[i]); */

        return 0;
}


int main(int argc, char *argv[])
{
	int  i;

	if (THREADS != 4) {
    		fprintf(stderr, "MUST BE RUN WITH 4 THREADS\n");
    		exit(1);
  	}

        if (argc != 4){
    		fprintf(stderr, "usage ./prog cache_line_size interval sync_type (0/1)\n");
    		exit(1);
  	}
        
        CACHE_LINE = atoi(argv[1]);
        INTERVAL = atoi(argv[2]);
        sync_type = atoi(argv[3]);

        N = (1.5*CACHE_LINE+INTERVAL)*THREADS;

        private_arr = (int *)malloc(THREADS*(1.5*CACHE_LINE+INTERVAL)*sizeof(int));
        if (private_arr == NULL){
	   	fprintf(stderr, "unable to allocate memory for local array\n");
        }
	global_arr_ptr = (shared[96] int *) upc_all_alloc(THREADS, N * sizeof(int));
	//global_arr_ptr = (shared[1] int *) upc_all_alloc(THREADS, N * sizeof(int));
	if (global_arr_ptr == NULL){
	   	fprintf(stderr, "unable to allocate memory using shared ptr\n");
	}
        start = global_arr_ptr;


	if (MYTHREAD == 0) {   /* initalizing the global_arr  */
		for (i = 0; i < 1.5 * CACHE_LINE + INTERVAL; i++) {
			*global_arr_ptr = 0;
			global_arr_ptr = global_arr_ptr + 1;
			private_arr[i] = 0;
		}

		for (i = 1.5*CACHE_LINE+INTERVAL; i < 3 * CACHE_LINE+2*INTERVAL; i++) {
			*global_arr_ptr = 10000;
			global_arr_ptr = global_arr_ptr + 1;
			private_arr[i] = 0;
                }
                

		for (i = 3*CACHE_LINE+2*INTERVAL; i < 4.5*CACHE_LINE+3*INTERVAL; i++) {
			*global_arr_ptr = 20000;
			global_arr_ptr = global_arr_ptr + 1;
			private_arr[i] = 0;
                }
		for (i = 4.5*CACHE_LINE+3*INTERVAL; i < N; i++) {
			*global_arr_ptr = 30000;
			global_arr_ptr = global_arr_ptr + 1;
			private_arr[i] = 0;
                }
		global_arr_ptr = start;
	}

	upc_barrier;

        /* allocating the locks for use in sync */
	lock = upc_all_lock_alloc();
	
        run_update_test();

	upc_barrier;

        if (MYTHREAD == 0) {
        	verify_updates();
        }
   
	upc_barrier;
       
        free(private_arr); 
        upc_free(global_arr_ptr);
        upc_lock_free(lock);

	printf("[%d] DONE\n", MYTHREAD);
     
        exit(0);
}
